"""user group viewer paths

Revision ID: da2cb2ac4255
Revises: 94a47ea2d3ff
Create Date: 2024-09-26 14:39:26.135101

"""
import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "da2cb2ac4255"
down_revision = "94a47ea2d3ff"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("usergroup", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "viewer_paths", sa.JSON(), server_default="[]", nullable=False
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("usergroup", schema=None) as batch_op:
        batch_op.drop_column("viewer_paths")

    # ### end Alembic commands ###
