"""Make ApplyWorkflow.start_timestamp not nullable

Revision ID: e75cac726012
Revises: 84bf0fffde30
Create Date: 2023-10-30 15:51:18.808789

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e75cac726012"
down_revision = "84bf0fffde30"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("applyworkflow", schema=None) as batch_op:
        batch_op.alter_column(
            "start_timestamp", existing_type=sa.DATETIME(), nullable=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("applyworkflow", schema=None) as batch_op:
        batch_op.alter_column(
            "start_timestamp", existing_type=sa.DATETIME(), nullable=True
        )

    # ### end Alembic commands ###
