"""Make ApplyWorkflow.workflow_dump non-nullable

Revision ID: d4fe3708d309
Revises: e75cac726012
Create Date: 2023-11-30 14:58:23.917824

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d4fe3708d309"
down_revision = "e75cac726012"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("applyworkflow", schema=None) as batch_op:
        batch_op.alter_column(
            "workflow_dump", existing_type=sa.JSON(), nullable=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("applyworkflow", schema=None) as batch_op:
        batch_op.alter_column(
            "workflow_dump", existing_type=sa.JSON(), nullable=True
        )

    # ### end Alembic commands ###
