HISTORY_FILENAME = "history.json"
FILTERS_FILENAME = "filters.json"
IMAGES_FILENAME = "images.json"
METADATA_FILENAME = "metadata.json"
SHUTDOWN_FILENAME = "shutdown"
WORKFLOW_LOG_FILENAME = "workflow.log"
